/*****************************************************************************
 *   Project:     GENETIC5.PRJ                                               *
 *****************************************************************************
 *   Modules:     GENPARM5.H                                                 *
 *                EXTERN5.H                                                  *
 *                INTERN5.H                                                  *
 *                GENETIC5.C                                                 *
 *                POPUL5.C                                                   *
 *                EVAL5.C                                                    *
 *                DISPLAY5.C                                                 *
 *                REPRO5.C                                                   *
 *****************************************************************************
 *   Topic:       EVAL5.C                                                     *
 *****************************************************************************
 *   Description: This module contains functions and procedures, that are    *
 *                responsible for the evaluation of one generation. The      *
 *                higher the value (fitness) the better the generation.      *
 *****************************************************************************/


#include <stdio.h>
#include <math.h>
#include <values.h>
#include <stdlib.h>
#include <string.h>
#include <mem.h>
#include "extern5.h"


/*****************************************************************************
 *   Function:    DECODE                                                     *
 *****************************************************************************
 *   Description: This function translates the binary coded value of an      *
 *                individuum into its real given range.                      *
 *****************************************************************************
 *   Input:       unsigned char  x - contains the uncoded x-value            *
 *****************************************************************************
 *   Return:      OPERAND_TYPE decode() - the translated value               *
 *****************************************************************************/
OPERAND_TYPE decode(unsigned char x)
{
    OPERAND_TYPE value;

    value  = LOWEST_OPERAND  + (LARGEST_OPERAND  - LOWEST_OPERAND)  * (OPERAND_TYPE)x / (OPERAND_TYPE)255;
    return(value);
}



/*****************************************************************************
 *   Function:    EVAL_FUNCTION                                              *
 *****************************************************************************
 *   Description: This function computes the fitness of one single           *
 *                chromosome. This is the real evaluation function.          *
 *****************************************************************************
 *   Input:       CHROMOSOME *ind - contains the binary coded individuum     *
 *****************************************************************************
 *   Output:      CHROMOSOME *ind - The fitness of the individuum is computed*
 *****************************************************************************
 *   Return:      double eval_function() - The fitness of the chromosome     *
 *****************************************************************************/

double eval_function(CHROMOSOME *ind)
{
  double value,
/*       limit, */
         result;            /* The fitness of the actual chromosome */
  int i;

  /* The quality of an individuum is simply defined as the difference between
     the real value of the function and the value of the generated function for each
     evaluation point */

/*  limit = sqrt(MAXDOUBLE); */
  for(i=0,value = 0.0;i<INPUT_ELEMENTS;++i)
  {
      result = dest_function((double)(i+1)) - function(ind,(double)(i+1)); /* error of one ordinate */
/*      if(result >= limit - value || result <= -limit + value)
      {
          value = MAXDOUBLE;
          break;
      }
      else
         value += result*result; */   /* Sum up the total error of all evaluation points */

     if(result < 0)
        result *= -1;

     if(value + result >= MAXDOUBLE)
        value = MAXDOUBLE;
     else
        value += result;
  }

  /* The closer the value to zero the better the result of the approximation.
     However, the internal logic of the GA is different: The larger the
     fitness the better the individuum. Therefore a translation has to be done
     realized with the reciprocal value. To avoid overflows or divisions by
     zero catch the value if nearly zero. */

  ind->fitness = (value > -1e-300 && value < 1e-300) ? MAXDOUBLE : (1.0 / value);
  if(ind->fitness > very_best_ind.fitness)
     memcpy(&very_best_ind,ind,sizeof(CHROMOSOME));
  return(ind->fitness);
}


/*****************************************************************************
 *   Function:    SORT_FUNCTION                                              *
 *****************************************************************************
 *   Description: This function is used by the quicksort-algorithm for       *
 *                comparison of two items.                                   *
 *****************************************************************************
 *   Input:       SORT_STRUCT *a - first item to be sorted                   *
 *                SORT_STRUCT *b - second item to be sorted                  *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      1 if a < b, -1 if a > b, 0 if equal                        *
 *****************************************************************************/

int sort_function(SORT_STRUCT *a, SORT_STRUCT *b)
{
  if(a->value < b->value)
     return(1);
  else
    if(a->value > b->value)
       return(-1);
    else
       return(0);
}


/*****************************************************************************
 *   Function:    FITNESS_IS_EVALUATION                                      *
 *****************************************************************************
 *   Description: fitness-technique that simply copies the fitness computed  *
 *                by the evaluation function to the individuum.              *
 *****************************************************************************
 *   Input:       GENERATION *g - not needed here                            *
 *                double p1     - not needed here                            *
 *                double p2     - not needed here                            *
 *****************************************************************************
 *   Output:      none                                                       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void fitness_is_evaluation(GENERATION *g,double p1,double p2)
{
     if(!str_fitness[0])
        strcpy(str_fitness,"fitness is evaluation");

/*     p1 = p1;
     p2 = p2;
     g  = g;    */
}


/*****************************************************************************
 *   Function:    WINDOWING                                                  *
 *****************************************************************************
 *   Description: fitness-technique that subtracts the fitness of the worst  *
 *                individuum from the firness of each other individuum.      *
 *                An optional parameter p1 can be specified as a minimum     *
 *                fitness that is assigned to each individuum with a lower   *
 *                fitness than p1.                                           *
 *****************************************************************************
 *   Input:       GENERATION *g - generation to be evaluated                 *
 *                double p1     - minimum fitness                            *
 *                double p2     - not needed here                            *
 *****************************************************************************
 *   Output:      GENERATION *g - generation with new fitness                *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void windowing(GENERATION *g,double p1,double p2)
{
    int i;                                  /* loop index */
    double value;                           /* total fitness of generation */

     if(!str_fitness[0])
        sprintf(str_fitness,"windowing with minimum of %lf",p1);
    p2 = p2;

    for(i=0,value = 0.0;i<POP_SIZE;++i)     /* compute fitness for all individuums */
    {
        /* subtract fitness of worst individuum */

        g->individuum[i].fitness = g->individuum[i].fitness - g->individuum[WORST].fitness;

        /* set to minimum fitness, if fitness is lower than minimum */

        if(g->individuum[i].fitness < p1)
           g->individuum[i].fitness = p1;

        /* sum up fitness of individuum to total fitness */

        value += g->individuum[i].fitness;
    }
    g->total_fitness = value;               /* total fitness of generation */
    update_sort_array(g);                   /* sort individuums by fitness */
}


/*****************************************************************************
 *   Function:    LINEAR_FITNESS                                             *
 *****************************************************************************
 *   Description: fitness-technique that assigns a given value p1 to the     *
 *                fittest individuum. All other individuums, sorted by       *
 *                fitness in decreasing order, are assigned a fitness that   *
 *                is p2 worse than the preceeding individuum.                *
 *****************************************************************************
 *   Input:       GENERATION *g - generation to be evaluated                 *
 *                double p1     - maximum fitness assigned to the best       *
 *                double p2     - value subtracted from p1                   *
 *****************************************************************************
 *   Output:      GENERATION *g - generation with new fitness                *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void linear_fitness(GENERATION *g,double p1,double p2)
{
    double value;                       /* total fitness of a generation */
    int i;                              /* loop index */

    if(!str_fitness[0])
       sprintf(str_fitness,"linear normalization with maximum %7.3lf and decrement %7.3lf",p1,p2);

    for(i=0,value = 0.0;i<POP_SIZE;++i)  /* compute fitness for all individuums */
    {
        /* sort_array[0] contains the position of the best individuum */
        /* assign p1 to the fittest individuum */

        g->individuum[sort_array[i].position].fitness = p1;

        /* if possible, subtract p2 from p1 for the next fittest individuum */
        /* otherwise set fitness to zero */

        p1 -= p2;
        if(p1 < 0)
            p1 = 0;

        /* sum up fitness of individuum to total fitness */

        value += g->individuum[sort_array[i].position].fitness;
    }
    g->total_fitness = value;               /* total fitness of generation */
    update_sort_array(g);                   /* sort individuums by fitness */
}


/*****************************************************************************
 *   Function:    UPDATE_SORT_ARRAY                                          *
 *****************************************************************************
 *   Description: this procedure creates an array, that contains the fitness *
 *                and number of each individuum of a generation, sorted by   *
 *                fitness in decreasing order. That means  that sort_array[0]*
 *                contains the fittest individuum, sort_array[POP_SIZE-1]    *
 *                the worst, respectively.                                   *
 *****************************************************************************
 *   Input:       GENERATION *g - generation to be sorted                    *
 *****************************************************************************
 *   Output:      sort_array    - contains individuums in sorted order       *
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void update_sort_array(GENERATION *g)
{
    int i;                                             /* loop index */
    typedef int (*fcmp) (const void *, const void *);  /* comparison function */

    for(i=0;i<POP_SIZE;++i)                            /* for all individuums */
    {
        /* fill array with fitness and number of individuums unsorted */

        sort_array[i].value = g->individuum[i].fitness;
        sort_array[i].position = i;
    }

    /* sort array by fitness via quicksort-algorithm */

    qsort((void *)sort_array, POP_SIZE, sizeof(SORT_STRUCT),(fcmp) sort_function);
}



/*****************************************************************************
 *   Function:    EVALUATE                                                   *
 *****************************************************************************
 *   Description: This function computes the fitness of the whole generation *
 *****************************************************************************
 *   Input:       GENERATION *gen                - the actual generation     *
 *                fitness_type fitness_technique - fitness technique to be   *
 *                                                 used for evaluation       *
 *                double parm1                   - parameter for fitness-    *
 *                                                 techniques                *
 *                double parm2                   - parameter for fitness-    *
 *                                                 techniques                *
 *****************************************************************************
 *   Output:      GENERATION *gen - the fitness of the generation is computed*
 *****************************************************************************
 *   Return:      none                                                       *
 *****************************************************************************/

void evaluate(GENERATION *gen,fitness_type fitness_technique,double parm1,double parm2)
{
    int i;                       /* index for the population */
    double value;                /* fitness of the generation */

    /* sum up the fitness of each individuum of the population */

    for(i=0, value = 0.0;i<POP_SIZE;++i)
        value += eval_function(&gen->individuum[i]);

    gen->total_fitness = value;          /* assign total fitness of generation */
    update_sort_array(gen);              /* sort indivisuums by fitness */
    fitness_technique(gen,parm1,parm2);  /* change fitness via given fitness-technique */
    fitness_chart[gen->number-1].best_fitness = gen->individuum[BEST].fitness;
    fitness_chart[gen->number-1].worst_fitness = gen->individuum[WORST].fitness;
    fitness_chart[gen->number-1].average_fitness = gen->total_fitness/POP_SIZE;
    display_result(gen);                /* display result on screen */
}
