/*****************************************************************************
 *   Project:     GENETIC5.PRJ                                               *
 *****************************************************************************
 *   Modules:     GENPARM5.H                                                 *
 *                EXTERN5.H                                                  *
 *                INTERN5.H                                                  *
 *                GENETIC5.C                                                 *
 *                POPUL5.C                                                   *
 *                EVAL5.C                                                    *
 *                DISPLAY5.C                                                 *
 *                REPRO5.C                                                   *
 *****************************************************************************
 *   Topic:       GENPARM5.H                                                 *
 *****************************************************************************
 *   Description: This module contains all global variables, structures,     *
 *                type definitions and parameters to control the GA.         *
 *****************************************************************************/

#define INPUT_ELEMENTS      15                      /* Number of evaluation points for comparison */
#define OPERATORS           10                      /* Number of operators to be used for function */
#define CHROMOSOME_SIZE     2*OPERATORS
#define POP_SIZE            100                     /* number of individuums per generation */
#define TRIALS              80000L                  /* desired size of total population */
#define ARRAY_SIZE          50                      /* number of items in destination function */
#define CROSS_OVER_RATE     0.8                     /* crossover probability */
#define MUTATION_RATE       0.04                    /* mutation probability */
#define REPLACE_RATE        0                       /* new population is replaced by the n best of the old generation */
#define ELITISM_RATE        10                      /* number of best individuums not to be reproduced */
#define MODIFY_ELITIST      1                       /* 1=mutate elitists
#define LOWEST_OPERAND      (double)0.0             /* lowest value of real alpha */
#define LARGEST_OPERAND     (double)10.0            /* largest value of real alpha */
#define OPERAND_TYPE        int                     /* type of the operands (int or double) */
#define FITNESS_TYPE &fitness_is_evaluation,0,0     /* desired fitness_techniqe for evaluation
                                                       available: - fitness_is_evaluation
                                                                  - windowing
                                                                  - linear_normalization        */
#define SELECTION_TYPE &roulette_wheel              /* desired parent selection
                                                       available: - roulette_wheel              */
#define BEST sort_array[0].position                 /* position of best individuum in generation */
#define WORST sort_array[POP_SIZE-1].position       /* position of worst individuum in generation */
#define XSTEP 0.1                                   /* steprate for function drawing */
#define MINX  1.0                                   /* first value to be graphically displayed */
#define MAXX  15.0                                  /* last value to be graphically displayed */
#define PAUSE_MODE 0                                /* 1=key must be pressed after evolution has stopped */
#define LOG_MODE 1                                  /* 1=creates logfile
#define AUTOSCALE 1                                 /* 1=for automatic scaling (function always fits on screen)
                                                       0=for static scaling (only original charts fits on screen) */
#define TRIGONOMETRIC 0                             /* 1 includes also trigonometric functions sinus and cosinus */
                                                    /* 0 only includes identity, reciprocal, increment, decrement */
                                                    /*   addition, exponential, squareroot, logarithm and negation */
typedef struct chromosome                           /* structure of a chromosome */
{
       double fitness;                              /* fitness of the chromosome */
       unsigned char coded_str[CHROMOSOME_SIZE];    /* odd positions contain operators, even positions operands */
} CHROMOSOME;


typedef struct generation                         /* structure of a generation */
{
       int number;                                /* number of the generation */
       double total_fitness;                      /* total fitness of the generation */
       CHROMOSOME individuum[POP_SIZE];           /* array of individuums */
} GENERATION;

typedef struct sort_struct                        /* structure of sorted individuums */
{
     double value;                                /* fitness of individuum */
     int position;                                /* position of individuum in generation */
} SORT_STRUCT;

typedef struct fitness_graph
{
     double best_fitness;
     double worst_fitness;
     double average_fitness;
} FITNESS_GRAPH;


/* function declarations */

typedef void (*selection_type)(CHROMOSOME *p,GENERATION *g);
typedef void (*fitness_type)(GENERATION *g,double p1,double p2);

enum { RECI, ADD, MULT, EXP, SIN, COS, SQRT, LN, NEG, POW, SUB, DIV, IDENT };
enum { TEXT, GRAPHIC, DETAILS, FITNESS, EXIT };

double eval_function(CHROMOSOME *ind);
double dest_function(double x);
OPERAND_TYPE decode(unsigned char x);
double function(CHROMOSOME *ind,double x);
void evaluate(GENERATION *gen,fitness_type fitness_techniqe,double parm1,double parm2);
void display_result(GENERATION *gen);
void graphic_display(GENERATION *gen);
void text_display(GENERATION *gen);
void detail_display(GENERATION *gen);
void fitness_display(int gen);
void write_file(GENERATION *gen);
void get_formula(char *txt,CHROMOSOME *ind);
void update_sort_array(GENERATION *gen);
void initialize(GENERATION *gen);
double polynome(CHROMOSOME *ind,double x);
void roulette_wheel(CHROMOSOME *p,GENERATION *g);
void replace(GENERATION *g1,GENERATION *g2,int count);
void populate(GENERATION *gen,selection_type parent_select,long trials);
void crossover(CHROMOSOME *c1, CHROMOSOME *c2);
void set_bit(unsigned char *value,int bit_no);
void clear_bit(unsigned char *value,int bit_no);
void mutate(CHROMOSOME *c,double mr);
void reproduce(CHROMOSOME *p1, CHROMOSOME *p2, CHROMOSOME *c1, CHROMOSOME *c2, double cr, double mr);
void fitness_is_evaluation(GENERATION *g,double p1,double p2);
void windowing(GENERATION *g,double p1,double p2);
void linear_fitness(GENERATION *g,double p1,double p2);
int sort_function(SORT_STRUCT *a, SORT_STRUCT *b);
int one_of_the_best(int position,int rate);
