/*  */
/* (c) 1993 by NEUROTEC GmbH 					         */
/* All rights reserved							 */
/*  */
/* Filename   : EVOLUT.C                                                 */
/* Author     : Sven Feddersen						 */
/* Project    : GenLib     						 */
/*  */
/*									 */
/* Publics :								 */
/*                                                                       */
/*   int _GenLibError;                                                   */
/*   int evolution(unsigned maxgenerations, unsigned popsize,            */
/*                 unsigned hold, boolean filtertwins,                   */
/*		   GLT_SelectionType select, GLT_CrossoverType cross,    */
/*		   double mutrate, double mutdim, GLT_MutationType mute, */
/*		   GLT_EvalType evalfun, GLT_ControlType control,        */
/*		   GLT_AllocType alloc, GLT_InitType init,               */
/*		   GLT_CopyType copy, GLT_FreeType forget);              */
/*                                                                       */
/*  */


#ifdef __TURBOC__
#include <stdlib.h>
#include <values.h>
#else
#include <stdlib.h>
#include <string.h>
#include <values.h>
#endif

//#include <compat.h>
#include <genlib.h>


int _GenLibError;

extern unsigned OrderLen;

static POPULATION *actpop;
static POPULATION *newpop;



static int compare(const void *idx1, const void *idx2) {
CHROM *c1 = &actpop->rep[*((unsigned *)idx1)];
CHROM *c2 = &actpop->rep[*((unsigned *)idx2)];
double fit1 = c1->fitness;
double fit2 = c2->fitness;
unsigned gen1, gen2;


  if (fit1 > fit2)
    return(-1);
  if (fit1 < fit2)
    return(1);
  gen1 = c1->myGeneration;
  gen2 = c2->myGeneration;
  if (gen1 > gen2)
    return(1);
  if (gen1 < gen2)
    return(-1); 
  return(0);
}



int evolution(unsigned popsize, unsigned hold,
	      boolean filtertwins, GLT_SelectionType select,
	      GLT_CrossoverType cross, double mutrate, double mutdim,
	      GLT_MutationType mute, GLT_EvalType evalfun,
	      GLT_ControlType control, GLT_AllocType alloc,
	      GLT_InitType init, GLT_CopyType copy, GLT_FreeType forget) {

POPULATION pop0, pop1, *memo;
int result, cont, generation = 0;
double worstFit, help;
CHROM huge *currChrom;
CHROM huge *parent1;
CHROM huge *parent2;
CHROM huge *child;
unsigned p1, p2, u, idx;
unsigned *ptrFit;


  pop0.size = pop1.size = popsize;

  pop0.fitvec = (unsigned *) malloc(popsize * sizeof(unsigned));
  if (!pop0.fitvec) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }

  pop1.fitvec = (unsigned *) malloc(popsize * sizeof(unsigned));
  if (!pop1.fitvec) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }

  do {
    if (!generation) {

      /* Aufbau und Initialisierung der beiden Populationen */
      /* nur in der 0. Generation                           */

      if (!alloc(popsize, &pop0)) return(0);
      if (!alloc(popsize, &pop1)) {
	forget(&pop0);
	return(0);
      }
      if (!init(&pop0)) return(0);

      parent1 = pop0.rep;
      parent2 = pop1.rep;
      for (u=0; u<popsize; u++, parent1++, parent2++) {
	parent1->mutprob = parent2->mutprob = mutrate;
	parent1->mutdim = parent2->mutdim = mutdim;
	parent1->fitness = evalfun((CHROM *) parent1);
        parent1->myGeneration = generation;
	pop0.fitvec[u] = pop1.fitvec[u] = u;
      }
      actpop = &pop0;
      newpop = &pop1;
    }
    else {
      memo = actpop;
      actpop = newpop;
      newpop = memo;
    }

    qsort(actpop->fitvec, popsize, sizeof(unsigned), compare);

    actpop->fitsum=0;
    actpop->invfitsum=0;

    worstFit = actpop->rep[actpop->fitvec[0]].fitness;
    for (u=0, currChrom=actpop->rep; u<popsize; u++, currChrom++) {
      actpop->fitsum += currChrom->fitness;
      help = worstFit - currChrom->fitness;
      actpop->invfitsum += max(help, MINDOUBLE);
    }

    cont = control(generation, actpop);

    if (cont) {
      for (u=0, ptrFit=actpop->fitvec; u<popsize; u++, ptrFit++) {
	idx = *ptrFit;

        /* Neue Kinder generieren */

        if (u<popsize-hold) {
	  child = newpop->rep;
	  child += idx;
	  p1 = select(actpop);
	  parent1 = parent2 = actpop->rep;
	  if (cross) {
	    p2 = select(actpop);
	    parent1 = parent2 = actpop->rep;
	    parent1 += p1;
	    parent2 += p2;
	    cross((CHROM *) parent1, (CHROM *) parent2, (CHROM *) child);
	    child->mutprob = (parent1->mutprob + parent2->mutprob) / 2.0;
	    child->mutdim = (parent1->mutdim + parent2->mutdim) / 2.0;
	  }
	  else {
	    parent1 = actpop->rep;
	    parent1 += p1;
	    copy((CHROM *) child, (CHROM *) parent1);
	  }
	  if (mute) {
	    result = mute((CHROM *) child);
	    if (!result) {
	      forget(&pop1);
	      forget(&pop0);
	      return(0);
	    }
	  }
	  newpop->rep[idx].fitness = evalfun(&newpop->rep[idx]);
          newpop->rep[idx].myGeneration = generation;
        }

        /* Eliten kopieren */

        else {
          copy(&newpop->rep[idx], &actpop->rep[idx]);
          newpop->rep[idx].fitness = evalfun(&newpop->rep[idx]);
        } 
      }
      generation++;
    }
  } while (cont);
  free(pop1.fitvec);
  free(pop0.fitvec);
  forget(&pop1);
  forget(&pop0);
  return(1);
}
