/*  */
/* (c) 1993 by NEUROTEC GmbH 					         */
/* All rights reserved							 */
/*  */
/* Filename   : ORDBASE.C                                                */
/* Author     : Sven Feddersen						 */
/* Project    : GenLib     						 */
/*  */
/*									 */
/* Publics :								 */
/*                                                                       */
/*  void OrderCopy(CHROM *d, CHROM *s);                                  */
/*  int OrderAlloc(unsigned popsize, POPULATION *pop);                   */
/*  int OrderRandInit(POPULATION *pop);                                  */
/*  int OrderFree(POPULATION *pop);                                      */
/*									 */
/*  */

#include <alloc.h>
#include <string.h>

#include <genlib.h>
#include <order.h>


extern int _GenLibError;


unsigned OrderLen = 1;


/*************************************************************************/
/*  OrderCopy kopiert ein Individuum, das in der Reihenfolge-            */
/*  Reprsentation kodiert ist.                                          */
/*                                                                       */
/*  d : Zielindividuum                                                   */
/*  s : Quellindividuum                                                  */
/*************************************************************************/

void OrderCopy(CHROM *d, CHROM *s) {

  memcpy(((OrderRep *) d->myrep)->job, ((OrderRep *) s->myrep)->job, ((OrderRep *) d->myrep)->n * sizeof(unsigned));
  d->fitness = s->fitness;
  d->mutprob = s->mutprob;
  d->mutdim = s->mutdim;
}



/*************************************************************************/
/*  OrderIsTwin prft, ob ein Individuum, das in der Reihenfolge-        */
/*  Reprsentation kodiert ist in einer Population bereits vorhanden     */
/*  ist.                                                                 */
/*                                                                       */
/*  c   : Individuum                                                     */
/*  pop : Population                                                     */
/*************************************************************************/

boolean OrderIsTwin(POPULATION *pop, CHROM *o) {
boolean twin=FALSE;
unsigned i;

  for (i=0; !twin && i<pop->size; i++)
    if (!memcmp(((OrderRep *) o->myrep)->job,
		((OrderRep *) pop->rep[i].myrep)->job,
		((OrderRep *) o->myrep)->n * sizeof(unsigned))) twin = FALSE;
  return(twin);
}



/*************************************************************************/
/*  OrderInit allokiert den Speicherplatz fr eine Population fr das    */
/*  Arbeiten auf der Reihenfolge-Reprsentation.                         */
/*                                                                       */
/*  popsize : Populationsgre                                           */
/*  pop     : Zeiger auf die beiden Populationen                         */
/*************************************************************************/

int OrderAlloc(unsigned popsize, POPULATION *pop) {
char *allmem, *actmem;
unsigned u;


  allmem = (char *) farmalloc((unsigned long) popsize *
	   (OrderLen * sizeof(unsigned) + sizeof(CHROM) + sizeof(OrderRep)));

  if (!allmem) {
    _GenLibError = GLE_NO_MEMORY;
    return(0);
  }
  actmem = allmem;

  pop->rep = (CHROM *) actmem;
  actmem += popsize * sizeof(CHROM);

  for (u=0; u<popsize; u++) {
    pop->rep[u].myrep = (OrderRep *) actmem;
    actmem += sizeof(OrderRep);
    ((OrderRep *) pop->rep[u].myrep)->n = OrderLen;
    ((OrderRep *) pop->rep[u].myrep)->job = (unsigned *) actmem;
    actmem += OrderLen * sizeof(unsigned);
  }
  return(1);
}



/*************************************************************************/
/*  OrderRandInit initialisiert die Populationen pop fr das Arbeiten    */
/*  auf der Reihenfolge-Reprsentation mit zuflligen Reihenfolgen.      */
/*                                                                       */
/*  popsize : Populationsgre                                           */
/*  pop     : Zeiger auf die Population                                  */
/*************************************************************************/

int OrderRandInit(POPULATION *pop) {
unsigned u;

  for (u=0; u<pop->size; u++) {
    EqualRandomIntVec(((OrderRep *) pop->rep[u].myrep)->job, OrderLen, OrderLen);
  }
  return(1);
}



/*************************************************************************/
/*  OrderFree gibt den von einer Populationen belegten Speicherplatz     */
/*  wieder frei.                                                         */
/*                                                                       */
/*  pop : Zeiger auf die Population                                      */
/*************************************************************************/

int OrderFree(POPULATION *pop) {

  farfree(pop);
  return(1);
}
