/*  */
/* (c) 1993 by NEUROTEC GmbH 					         */
/* All rights reserved							 */
/*  */
/* Filename   : TRAVEL.C                                                 */
/* Author     : Sven Feddersen						 */
/* Project    : TSP        						 */
/*  */
/*									 */
/*  TSP fr folgende Stdte :                                            */
/*                                                                       */
/*  00: Aachen                                                           */
/*  01: Berlin                                                           */
/*  02: Bielefeld                                                        */
/*  03: Braunschweig                                                     */
/*  04: Bremen                                                           */
/*  05: Dortmund                                                         */
/*  06: Dresden                                                          */
/*  07: Dsseldorf                                                       */
/*  08: Essen                                                            */
/*  09: Frankfurt/Main                                                   */
/*  10: Freiburg i. Br.                                                  */
/*  11: Hamburg                                                          */
/*  12: Hannover                                                         */
/*  13: Karlsruhe                                                        */
/*  14: Kassel                                                           */
/*  15: Kiel                                                             */
/*  16: Kln                                                             */
/*  17: Leipzig                                                          */
/*  18: Magdeburg                                                        */
/*  19: Mnchen                                                          */
/*  20: Nrnberg                                                         */
/*  21: Saarbrcken                                                      */
/*  22: Stuttgart                                                        */
/*  23: Wiesbaden                                                        */
/*  24: Wuppertal                                                        */
/*  */


#include <dos.h>
#include <time.h>
#include <ctype.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#include <genlib.h>
#include <order.h>

#define NUM_TOWNS 25         /* Anzahl der Staedte im TSP */

extern unsigned OrderLen;    /* Laenge der Reihenfolge im genetischen Code */



/* Es folgen die Zeilen der Abstandstabelle.             */
/* Wegen der Symmetrie genuegt die obere Dreiecksmatrix. */

int line00[] = { 633, 257, 412, 378, 150, 659,  80, 134, 259, 505,
		 488, 353, 372, 330, 581, 70, 585, 496, 648, 475,
		 246, 446, 235, 121 };

int line01[] = { 390, 227, 397, 488, 205, 572, 530, 555, 819, 289,
		 282, 686, 378, 370, 569, 179, 147, 584, 437, 745,
		 624, 581, 518 };

int line02[] = { 169, 175, 112, 511, 196, 154, 272, 618, 262, 110,
		 485, 119, 355, 191, 437, 253, 598, 428, 472, 559,
		 348, 142 };

int line03[] = { 167, 267, 304, 351, 309, 338, 602, 196, 61, 469,
		 150, 289, 346, 196,  85, 625, 452, 528, 543, 364,
		 297 };

int line04[] = { 233, 470, 317, 275, 466, 730, 119, 125, 597, 278,
		 278, 312, 362, 251, 753, 580, 656, 671, 469, 263 };

int line05[] = { 607, 63, 34, 264, 509, 343, 208, 397, 174, 436,
		  83, 532, 351, 653, 480, 364, 451, 240,  35 };

int line06[] = { 629, 638, 469, 728, 502, 364, 578, 389, 571, 589,
		 108, 236, 484, 337, 659, 524, 495, 636 };

int line07[] = {  29, 232, 478, 427, 292, 345, 276, 520,  47, 558,
		 435, 621, 448, 332, 419, 208, 29 };

int line08[] = { 249, 494, 385, 250, 361, 207, 478,  68, 567, 393,
		 638, 465, 349, 436, 225, 36 };

int line09[] = { 276, 495, 352, 143, 193, 588, 189, 395, 427, 395,
		 222, 202, 217,  32, 236 };

int line10[] = { 759, 616, 135, 457, 852, 435, 659, 691, 412, 391,
		 279, 207, 283, 482 };

int line11[] = { 154, 626, 307,  93, 422, 391, 280, 782, 609, 685,
		 700, 521, 373 };

int line12[] = { 483, 164, 247, 287, 256, 145, 639, 466, 542, 557,
		 378, 238 };

int line13[] = { 324, 719, 302, 526, 558, 287, 241, 144,  82, 150,
		 349 };

int line14[] = { 400, 249, 318, 239, 482, 309, 383, 398, 219, 222 };

int line15[] = { 515, 484, 373, 875, 702, 778, 793, 614, 466 };

int line16[] = { 515, 430, 578, 405, 289, 376, 165,  55 };

int line17[] = { 128, 438, 291, 591, 478, 427, 568 };

int line18[] = { 523, 376, 618, 563, 454, 381 };

int line19[] = { 167, 431, 220, 429, 625 };

int line20[] = { 395, 187, 254, 452 };

int line21[] = { 226, 165, 336 };

int line22[] = { 224, 423 };

int line23[] = { 212 };

int *lineno[24];


/* Die Namen der 25 Staedte */

char *names[] = {
  "Aachen         ", "Berlin         ", "Bielefeld      ", "Braunschweig   ",
  "Bremen         ", "Dortmund       ", "Dresden        ", "Dsseldorf     ",
  "Essen          ", "Frankfurt/Main ", "Freiburg i. Br.", "Hamburg        ",
  "Hannover       ", "Karlsruhe      ", "Kassel         ", "Kiel           ",
  "Kln           ", "Leipzig        ", "Magdeburg      ", "Mnchen        ",
  "Nrnberg       ", "Saarbrcken    ", "Stuttgart      ", "Wiesbaden      ",
  "Wuppertal      "
};

unsigned bestfit = MAXINT;
unsigned lastgen;


/*************************************************************************/
/* Die Bewertungsfunktion fuer das TSP. Fitness ist einfach die          */
/* kumulierte Wegstrecke.                                                */
/*************************************************************************/

double evaluate(CHROM *c) {
OrderRep *o = (OrderRep *) c->myrep;
unsigned act, lst, i;
double sum = 0;

  for (i=0; i<o->n; i++) {
    act = o->job[i];
    lst = (i ? o->job[i-1] : o->job[o->n-1]);
    if (act<lst) SWAP(act, lst);
    sum += lineno[lst][act-lst-1];
  }
  return(sum);
}



/*************************************************************************/
/* Die Kontrollfunktion wird nach jeder Generation einmal aufgerufen.    */
/* Es wird die Generationsnummer, die Fitness des besten und des         */
/* schlechtesten Individuums der Generation sowie die mittlere Fitness   */
/* ausgegeben. Wurde eine bessere Loesung gefunden, so wird auch diese   */
/* ausgegeben.                                                           */
/*************************************************************************/

int control(unsigned generation, POPULATION *pop) {
CHROM *theBest = pop->rep + pop->fitvec[pop->size-1];
CHROM *theWorst = pop->rep;
char s[20], c;
double sum=0.0;
OrderRep *o;
int i;

  lastgen = generation;
  if (theBest->fitness < bestfit) {
    o = (OrderRep *) theBest->myrep;
    for (i=0; i<NUM_TOWNS; i++) {
      gotoxy(1 + (i % 5) * 16, 9 + i / 5);
      printf(names[o->job[i]]);
    }
    bestfit = theBest->fitness;
  }

  for (i=0; i<pop->size; i++)
    sum += pop->rep[i].fitness;

  gotoxy(16, 1);
  printf("%6u", generation);
  gotoxy(16, 2);
  printf("%6u", (unsigned) theBest->fitness);
  gotoxy(16, 3);
  printf("%6u", (unsigned) theWorst->fitness);
  gotoxy(16, 4);
  printf("%6u", (unsigned) (sum/pop->size));

  if (kbhit()) {
    c = toupper(getch());
    switch (c) {
      case 27 : return(0);
    }
  }

  if (generation == 200)
    return(0);

  return(1);
}



void main(void) {
union REGS reg;
int i;


  OrderLen = NUM_TOWNS;
  randomize();

  lineno[0] = line00;
  lineno[1] = line01;
  lineno[2] = line02;
  lineno[3] = line03;
  lineno[4] = line04;
  lineno[5] = line05;
  lineno[6] = line06;
  lineno[7] = line07;
  lineno[8] = line08;
  lineno[9] = line09;
  lineno[10] = line10;
  lineno[11] = line11;
  lineno[12] = line12;
  lineno[13] = line13;
  lineno[14] = line14;
  lineno[15] = line15;
  lineno[16] = line16;
  lineno[17] = line17;
  lineno[18] = line18;
  lineno[19] = line19;
  lineno[20] = line20;
  lineno[21] = line21;
  lineno[22] = line22;
  lineno[23] = line23;

  clrscr();

  printf("Generation:\n");
  printf("Bester:\n");
  printf("Schlechtester:\n");
  printf("Mittel:\n\n\n");
  printf("Lsung:");

  evolution(40, 4, FALSE, InvRouletteSelection, EdgeRecombination, 0.005, 5, AdaptedMoveMutation, evaluate, control, OrderAlloc, OrderRandInit, OrderCopy, OrderFree);

  getch();
  clrscr();
}